--[[
	Project: SoulEnergyTracker
	Friendly Name: Soul Energy Tracker
	Author: Vandesdelca32

	File: core.lua
	Purpose: The main functions of the addon.

	Version: 8

	ALL RIGHTS RESERVED.
	COPYRIGHT (c)2015 VANDESDELCA32
]]

-- Import Engine:
local E, L, DF = unpack(select(2, ...))

-- GLOBALS: SoulEnergyTrackerData
local tostringall, tonumber, format, error = tostringall, tonumber, format, error
local strlower, strsplit, floor, strmatch, string = strlower, strsplit, floor, strmatch, string
local UnitName, UnitIsPlayer, random, GetRealmName = UnitName, UnitIsPlayer, random, GetRealmName
local time, C_Timer, pcall, loadstring = time, C_Timer, pcall, loadstring
local Ambiguate, SecondsToTime, tconcat = Ambiguate, SecondsToTime, table.concat

local update_message = false
E.Cache = {}

E.Commands = {
	-- This is where the chat commands are.
	["add"] = function(self, args)
		self.printFuncName("add", self, tconcat(args, ";"))
		local val = tonumber(args[2])
		if val then
			local newval = self.CurrentValue + val

			if newval < 0 then
				self:PrintErr(L["Err/ValOOR"])
				newval = 0
			elseif newval > SoulEnergyTrackerData.Maximum then
				self:PrintErr(L["Err/ValOOR"])
				newval = SoulEnergyTrackerData.Maximum
			end

			SoulEnergyTrackerData.Last_Value = newval
			self:PrintMessage(format(L["Msg/NewSoulEnergy"], SoulEnergyTrackerData.Last_Value))
			self:StartUpdating(true)
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
	end,

	["set"] = function(self, args)
		self.printFuncName("set", self, tconcat(args, ";"))
		local val = tonumber(args[2])
		if val then
			if val > SoulEnergyTrackerData.Maximum or val < 0  then
				self:PrintErr(format(L["Err/ValOORSpecific"], 0, SoulEnergyTrackerData.Maximum))
			end

			SoulEnergyTrackerData.Last_Value = val
			self:PrintMessage(format(L["Msg/NewSoulEnergy"], SoulEnergyTrackerData.Last_Value))
			self:StartUpdating(true)
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
	end,

	["setmax"] = function(self, args)
		self.printFuncName("setmax", self, tconcat(args, ";"))
		local val = tonumber(args[2])
		if val then
			if val < self.CurrentValue then
				self:PrintErr(L["Err/MaxTooLow"])
				return
			end

			SoulEnergyTrackerData.Maximum = val
			self:PrintMessage(format(L["Msg/Maximum"] .. L["Msg/NewSoulEnergy"], SoulEnergyTrackerData.Maximum))
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
	end,

	["setauto"] = function(self, args)
		self.printFuncName("setauto", self, tconcat(args, ";"))
		local val = tonumber(args[2])
		if val then
			if val <= 0 or val >= 1 then
				self:PrintErr(format(L["Err/ValOORSpecific"], 0, 1))
			else
				SoulEnergyTrackerData.Decline.Auto_Rate = val
				self:PrintMessage(format(L["Msg/AutoDecline"], SoulEnergyTrackerData.Decline.Auto_Rate * 100))
			end
		end
	end,

	["stop"] = function(self, args)
		self.printFuncName("stop", self, tconcat(args, ";"))
		self:StopUpdating()
		SoulEnergyTrackerData.Last_Value = self.CurrentValue
	end,

	["start"] = function(self, args)
		self.printFuncName("start", self, tconcat(args, ";"))
		self:StartUpdating(true)
	end,

	["hide"] = function(self, args)
		self.printFuncName("hide", self, tconcat(args, ";"))
		self._ShowBar = false
	end,

	["show"] = function(self, args)
		self.printFuncName("show", self, tconcat(args, ";"))
		self._ShowBar = true
	end,

	["decline"] = function(self, args)
		self.printFuncName("decline", self, tconcat(args, ";"))
		local rate, int
		if args[2] == "auto" then
			-- We'll calcuate the rate for the player here:
			local modifier = random(50, 150)
			int = 1440
			rate = (SoulEnergyTrackerData.Decline.Auto_Rate * (modifier / 100)) * SoulEnergyTrackerData.Maximum
			self:PrintMessage(format(L["Msg/SetDecline_Auto"],  ((100 * SoulEnergyTrackerData.Decline.Auto_Rate) * (modifier / 100))))
		else
			rate, int = tonumber(args[2]), tonumber(args[3])
		end

		if rate and int then
			if rate == 0 then
				self:PrintErr(L["Err/NoZeros"])
				return
			end
			if int <= 0 then
				self:PrintErr(format(L["Err/BiggerThan"], 0))
				return
			end
			-- Store the current value, so we don't lose it
			self.Commands["set"](self, {"set", self.CurrentValue})

			SoulEnergyTrackerData.Decline.Interval = int
			SoulEnergyTrackerData.Decline.Amount = rate


			-- let the player know what's up
			self:PrintMessage(format(L["Msg/SetDecline_Manual"],
			SoulEnergyTrackerData.Decline.Amount, SecondsToTime(SoulEnergyTrackerData.Decline.Interval * 60, nil, true)))
			-- self:StartUpdating(true)
		elseif args[2] == "get" then
			self:PrintMessage(format(L["Msg/SetDecline_Manual"],
			SoulEnergyTrackerData.Decline.Amount, SecondsToTime(SoulEnergyTrackerData.Decline.Interval * 60, nil, true)))
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
	end,

	["import"] = function(self, args)
		self.printFuncName("import", self, tconcat(args, ";"))
		if WeakAuras then
			StaticPopup_Show("SOULENERGY_WEAKAURASIMPORT1")
		end
	end,


	["debug"] = function(self, args)
		self.printFuncName("debug", self, tconcat(args, ";"))
		if not args[2] then args[2] = "toggle" end

		if args[2] == "on" then
			self._DebugMode = true
		elseif args[2] == "off" then
			self._DebugMode = false
		elseif args[2] == "toggle" then
			self._DebugMode = not self._DebugMode
		else
			self:PrintErr(L["Err/ArgsInvalid"])
		end
		self:PrintMessage("Debug Mode " .. ((self._DebugMode and "§aon§r.") or (not self._DebugMode and "§coff§r.")))
	end
}


--- Check the settings:
function E:CheckSettings()
	self.printFuncName("CheckSettings")
	if not SoulEnergyTrackerData then
		SoulEnergyTrackerData = DF
		self:PrintErr(L["Err/SettingsNotFound"])
	elseif SoulEnergyTrackerData._Version ~= DF._Version then
		SoulEnergyTrackerData = DF
	end
end

--- Gets the current value for Soul Energy based on some simple math.
--@return the current value for Soul Energy
function E:GetCurrentValue(target)
	self.printFuncName("GetCurrentValue", target)
	local actualValue -- This is the value this function will return
	local search = "Data"

	-- check to see if there's a target...
	if target then
		search = ".Cache." .. target
	end

	local code = [[
	if SoulEnergyTracker%s._LastChange == -1 then
		return SoulEnergyTracker%s.Last_Value
	else
		local timePassed = time() - SoulEnergyTracker%s._LastChange

		-- figure out how much to subtract from the time difference
		local totalDecline = floor((timePassed / (SoulEnergyTracker%s.Decline.Interval * 60) * SoulEnergyTracker%s.Decline.Amount))
		actualValue = SoulEnergyTracker%s.Last_Value - totalDecline
		return actualValue
	end
	]]

	code = string.gsub(code, "%%s", search)
	local success, rv = pcall(loadstring(code))
	if success then
		return rv
	else
		error(rv)
	end

end


--- Updates public values that other addons can use.
function E:RunUpdate(target)
	self.printFuncName("RunUpdate", target)
	-- Make the current value public
	if target then
		self.Cache[target].CurrentValue = self:GetCurrentValue(target)
		return
	end

	self.CurrentValue = self:GetCurrentValue()
	local name = UnitName("target")
	if self.Cache[name] then self:RunUpdate(name)
	end


	-- Check the ranges, to see if we have to stop.
	if self.CurrentValue > SoulEnergyTrackerData.Maximum then
		self.CurrentValue = SoulEnergyTrackerData.Maximum
		self:StopUpdating()
	elseif self.CurrentValue < 0 then
		self.CurrentValue = 0
		self:StopUpdating()
	end
end


function E:HandleChatCommand( rawargs )
	self.printFuncName("HandleChatCommand", rawargs)
	local args = {strsplit(" ", strlower(rawargs))}
	if self.Commands[args[1]] then
		self.Commands[args[1]](self, args)
	else
		self:PrintErr(L["Err/ArgsInvalid"])
	end
end

function E:StartUpdating( updateTime )
	self.printFuncName("StartUpdating", updateTime)
	if updateTime then
		SoulEnergyTrackerData._LastChange = time()
	end

	-- Do not create a new timer if there is one already running;
	-- just force an update.
	self:RunUpdate()
	if self.ticker then
		return
	end

	self._ShowBar = true
	self.ticker = C_Timer.NewTicker(SoulEnergyTrackerData.Update_Interval, function() self:RunUpdate() end)
end

function E:StopUpdating()
	self.printFuncName("StopUpdating")
	SoulEnergyTrackerData._LastChange = -1
	if self.ticker then
		self.ticker:Cancel()
		self:PrintMessage(L["Msg/Stopped"])
		self.ticker = nil
	else
		self:PrintErr(L["Msg/NotRunning"])
	end
end

function E:PLAYER_TARGET_CHANGED()
	self.printFuncName("PLAYER_TARGET_CHANGED")
	self:SendComm("Update", "target")
end

function E:AmbiguateName(target)
	self.printFuncName("AmbiguateName", target)
	local name, realm = UnitName(target)
	local fullName

	if not realm or realm == "" then
		realm = GetRealmName():gsub("%s+", "")
	end
	if name and realm then
		fullName = name .. "-" .. realm
		self:PrintDebug("  == " ..fullName)
		return Ambiguate(fullName, "none")
	end

end



function E:CheckVersion(versionData)
	self.printFuncName("CheckVersion", versionData)
	local newMinor = tonumber(strmatch(versionData, "rev(%d+)"))
	if not update_message and newMinor then
		if tonumber(newMinor) > self._Minor then
			update_message = true
			self:PrintErr(L["Msg/UpdateMinor"])
		end
	end
	self:PrintDebug(format("--- §9VerCheck: §dlocal§r=%d §eremote§r=%d", self._Minor, newMinor))
end


function E:OnCommReceived(prefix, message, scope, sender)
	self.printFuncName("OnCommReceived", prefix, message, scope, sender)
	-- Check what the message says: If we got an Update message, then we can respond with our settings.
	local messageType, messageData = strmatch(message, "(.+):(.+)")
	--
	if messageType == "UPDATEME" then
		self:CheckVersion(messageData)
		self:SendComm("Version", sender)
		self:SendComm("SendData", sender)
	elseif messageType == "DATA" then
		-- Write the data to the targetcache.
		local success, data = self:Deserialize(messageData)
		if success then
			self.Cache[sender] = data
			-- Update the bar, and get a current value
			self:RunUpdate(sender)
		else
			error("could not deserialze: " ..  data)
		end
	elseif messageType == "VER" then
		self:CheckVersion(messageData)
	end
end


function E:SendComm(messageType, target)
	self.printFuncName("SendComm", messageType, target)

	-- We don't send messages to NPCs...
	if not UnitIsPlayer(target) then return end

	-- Figure out who we're sending it to...
	local name = self:AmbiguateName(target)

	-- Message types are "Update", "SendData", and "Version"
	if messageType == "Update" then
		self:SendCommMessage(E._CommPrefix, "UPDATEME:" .. "rev" .. self._Minor, "WHISPER", name)
	elseif messageType == "SendData" then
		local SendData = self:Serialize(SoulEnergyTrackerData)
		self:SendCommMessage(E._CommPrefix, "DATA:" .. SendData, "WHISPER", name)
	elseif messageType == "Version" then
		self:SendCommMessage(E._CommPrefix, "VER:" .. "rev" .. self._Minor, "WHISPER", name)
	end
end

